from django.db import models
from django.conf import settings
from shortuuid import ShortUUID


class Url(models.Model):
    full_url = models.CharField(max_length=2083)
    short_id = models.CharField(max_length=200)
    redirects = models.IntegerField(default=0)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __str__(self):
        return "%s, %d" % (self.full_url, self.redirects)

    @staticmethod
    def gen_short_id():
        """
        Helper which randomly generates base62 id
        """
        return ShortUUID().random(length=settings.SHORT_ID_LEN)



from django.shortcuts import get_object_or_404
from django.http import HttpResponseRedirect,  Http404
from django.urls import reverse
from django.views import generic
from .models import Url


URLS_PAGINATION = 10
MOST_POPULAR_COUNT = 20


class IndexView(generic.ListView):
    template_name = 'url_simplify/index.html'
    context_object_name = 'popular_url_list'

    def get_queryset(self):
        """
        Return the most popular `MOST_POPULAR_COUNT` objects,
        based on redirects and created_at
        """
        return get_most_popular()[:MOST_POPULAR_COUNT]


class DetailView(generic.DetailView):
    model = Url
    template_name = 'url_simplify/detail.html'


class UrlsView(generic.ListView):
    template_name = 'url_simplify/urls_list.html'
    context_object_name = 'url_list'
    paginate_by = URLS_PAGINATION

    def get_queryset(self):
        """
        Pass most popular urls based on `get_most_popular()`
        """
        return get_most_popular()


def short_redirect(request, short_id):
    """
    Handles redirects for shortened urls, every time user request this,
    force redirect and increment redirect count for particular `Url` object
    """
    url = get_object_or_404(Url, short_id=short_id)
    url.redirects += 1
    url.save()
    redirect_url = url.full_url
    if not redirect_url.startswith('http://'):
        if not redirect_url.startswith('https://'):
            redirect_url = 'http://' + redirect_url
    return HttpResponseRedirect(redirect_url)


def add_url(request):
    """
    Url shortener form action
    """
    input_url = request.POST.get('input_url', '')
    # in case 'input_url' is not provided by POST request
    if not input_url:
        return HttpResponseRedirect(reverse('url_simplify:index'))
    short_id = Url.gen_short_id()
    # check if already exists, if so, try one to generate one more time
    tmp = Url.objects.filter(short_id=short_id)
    if tmp:
        return Http404("Something went wrong. Please, try one more time")
    url = Url(full_url=input_url, short_id=short_id)
    url.save()
    return HttpResponseRedirect(reverse('url_simplify:detail',
                                        args=(url.id,)))


def delete_url(request, url_id):
    """
    Url remove handler
    """
    p = Url.objects.get(pk=url_id)
    p.delete()
    return HttpResponseRedirect(reverse('url_simplify:urls'))


def get_most_popular():
    """
    Helper which retrieves most popular urls based on `redirects`(desc) and
    `created_at`(desc)
    """
    return Url.objects.order_by('-redirects', '-created_at')



from django.urls import path
from django.conf import settings
from . import views

app_name = 'url_simplify'
urlpatterns = [
    path('', views.IndexView.as_view(), name='index'),
    path('<int:pk>/', views.DetailView.as_view(), name='detail'),
    path('<int:url_id>/delete/', views.delete_url, name='delete_url'),
    path('urls/', views.UrlsView.as_view(), name='urls'),
    path('shortify/', views.add_url, name='add_url'),
    path(settings.SHORT_PREFIX + '<str:short_id>/',
         views.short_redirect, name='short_redirect')
]



from django.test import TestCase
from django.urls import reverse
from django.conf import settings
from .models import Url
from .views import MOST_POPULAR_COUNT


def create_url(full_url, redirects=0):
    """
    Create a url with the given `full_url` and optional `redirects`
    numbers.
    """
    return Url.objects.create(full_url=full_url, redirects=redirects)


class UrlModelTests(TestCase):

    def test_url_create(self):
        """
        test url create process
        """
        create_url(full_url="example.com")
        self.assertIs(len(Url.objects.all()), 1)

    def test_generate_short_id(self):
        """
        gen_short_id() returns randomly generated base62 id with
        length equals to ` settings.SHORT_ID_LEN`
        """
        short_id = Url.gen_short_id()
        self.assertIs(len(short_id), settings.SHORT_ID_LEN)


class UrlIndexViewTests(TestCase):

    def test_no_urls(self):
        """
        If no urls exist, an appropriate message is displayed.
        """
        response = self.client.get(reverse('url_simplify:index'))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "No simplified urls are available.")
        self.assertQuerysetEqual(response.context['popular_url_list'], [])

    def test_one_url(self):
        """
        Url displayed on the index page.
        """
        create_url(full_url="example.com")
        response = self.client.get(reverse('url_simplify:index'))
        self.assertQuerysetEqual(
            response.context['popular_url_list'],
            ['<Url: example.com, 0>']
        )

    def test_urls_order(self):
        """
        Url displayed based on `redirects` count and `created_at`
        on the index page.
        """
        create_url(full_url="example.com", redirects=20)
        create_url(full_url="example1_second_created.com", redirects=10)
        create_url(full_url="example1_last_created.com", redirects=10)
        response = self.client.get(reverse('url_simplify:index'))
        self.assertQuerysetEqual(
            response.context['popular_url_list'],
            ['<Url: example.com, 20>',
             '<Url: example1_last_created.com, 10>',
             '<Url: example1_second_created.com, 10>']
        )

    def test_most_popular_full_url(self):
        """
        Max number of most popular urls displayed on the index page.
        """
        for x in range(MOST_POPULAR_COUNT):
            create_url(full_url="example.com")
        response = self.client.get(reverse('url_simplify:index'))
        self.assertQuerysetEqual(
            response.context['popular_url_list'],
            ['<Url: example.com, 0>' for n in range(MOST_POPULAR_COUNT)]
        )

    def test_most_popular_plus_one_url(self):
        """
        Max number of most popular urls displayed on the index page.
        Even though there are max+1 urls.
        """
        for x in range(MOST_POPULAR_COUNT+1):
            create_url(full_url="example.com")
        response = self.client.get(reverse('url_simplify:index'))
        self.assertQuerysetEqual(
            response.context['popular_url_list'],
            ['<Url: example.com, 0>' for n in range(MOST_POPULAR_COUNT)]
        )


class UrlDetailViewTests(TestCase):

    def test_empty_url(self):
        """
        The url which has not created yet, should
        returns a 404 not found.
        """
        url = reverse('url_simplify:detail', args=(1,))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)
